/*
 * I2C support for the JZ4730.
 *
 * Copyright (C) 2017, 2018, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/i2c-jz4730.h>
#include <l4/devices/hw_mmio_register_block.h>

#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/util/util.h>

#include <cstdio>

/*
I2C pins are dedicated to I2C only and are not GPIO-controlled:

I2C0: Y4/SMB0_SDA,        V5/SMB0_SCK

Note that there is effectively only one I2C channel.
*/

enum Regs
{
  I2c_data                    = 0x000,  // I2CDR
  I2c_control                 = 0x004,  // I2CCR
  I2c_status                  = 0x008,  // I2CSR
  I2c_clock                   = 0x00c,  // I2CGR
};

enum I2c_control_bits : unsigned
{
  I2c_control_enable_irq      = 0x10,   // IEN
  I2c_control_start           = 0x08,   // STA
  I2c_control_stop            = 0x04,   // STO
  I2c_control_nack            = 0x02,   // AC
  I2c_control_enable          = 0x01,   // I2CE
};

enum I2c_status_bits : unsigned
{
  I2c_status_buffer_nempty    = 0x10,   // STX
  I2c_status_busy             = 0x08,   // BUSY
  I2c_status_transmit_end     = 0x04,   // TEND
  I2c_status_data_valid       = 0x02,   // DRF
  I2c_status_nack             = 0x01,   // ACKF
};

enum I2c_clock_values : unsigned
{
  I2c_clock_max       = 0xffff,
  I2c_clock_min       = 0,
};



// Initialise a channel.

I2c_jz4730_channel::I2c_jz4730_channel(l4_addr_t start,
                                       Cpm_jz4730_chip *cpm,
                                       uint32_t frequency,
                                       l4_cap_idx_t irq)
: _cpm(cpm), _frequency(frequency), _irq(irq)
{
  _regs = new Hw::Mmio_register_block<32>(start);
}

// Enable the channel.

void
I2c_jz4730_channel::enable()
{
  // Make sure that the I2C clock is available.

  _cpm->start_i2c();

  // Set the bus clock frequency.

  set_frequency();

  // Enable the channel and interrupts.

  _regs[I2c_control] = I2c_control_enable | I2c_control_enable_irq;
  while (!(_regs[I2c_control] & I2c_control_enable));
}

// Disable the channel.

void
I2c_jz4730_channel::disable()
{
  _regs[I2c_control] = 0;
  while (_regs[I2c_control] & I2c_control_enable);
}

// Set the frequency-related peripheral parameters.

void
I2c_jz4730_channel::set_frequency()
{
  // The APB clock (PCLK) is used to drive I2C transfers. Its value must be
  // obtained from the CPM unit and is scaled to kHz in order to keep the
  // numbers easily representable, as is the bus frequency.

  uint32_t pclk = _cpm->get_pclock_frequency() / 1000;
  uint32_t i2c_clk = _frequency / 1000;
  uint32_t division = pclk / (16 * i2c_clk);

  if (division > I2c_clock_min)
  {
    division -= 1;
    if (division > I2c_clock_max)
      division = I2c_clock_max;
  }

  _regs[I2c_clock] = division;
}

// Present the address on the bus.

bool
I2c_jz4730_channel::set_address(uint8_t address, bool read)
{
  // Waiting for long enough may eliminate a busy condition and thus permit a
  // new transaction. 10ms appears to be long enough, whereas 1ms does not
  // appear to be. In case this is insufficient, permit failure.

  unsigned int limit = 10;

  do
  {
    if (!wait_for_irq(1000) && !(--limit))
      return false;
  }
  while (busy());

  start();

  _regs[I2c_data] = (address << 1) | (read ? 1 : 0);

  send_next();

  return true;
}

// Wait up to the given timeout (in microseconds) for an interrupt request,
// returning true if one was delivered.

bool
I2c_jz4730_channel::wait_for_irq(unsigned int timeout)
{
  return !l4_error(l4_irq_receive(_irq, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4util_micros2l4to(timeout))));
}

// Read data from the bus.

unsigned int
I2c_jz4730_channel::read(uint8_t address, uint8_t buf[], unsigned int length)
{
  unsigned int nread = 0;

  if (!set_address(address, true))
    return 0;

  // Wait for an opportunity to begin reading.

  do
  {
    if (!wait_for_irq(1000000))
      printf("start (no irq): status = %x\n", (uint32_t) _regs[I2c_status]);
  }
  while (transferring() || (!data_valid() && !nack()));

  // Device apparently unavailable.

  if (nack())
  {
    stop();
    return nread;
  }

  // Attempt to read from the device.

  while (nread < length)
  {
    do
    {
      if (!wait_for_irq(1000000))
      {
        stop();
        return nread;
      }
    }
    while (!data_valid() && !nack());

    if (nack())
      break;

    if ((!nread && (length == 1)) || (nread == length - 2))
      signal_last();

    buf[nread++] = _regs[I2c_data];
    clear_next();
  }

  stop();
  return nread;
}

// Write data to the bus.

unsigned int
I2c_jz4730_channel::write(uint8_t address, uint8_t buf[], unsigned int length)
{
  unsigned int nwritten = 0;

  if (!set_address(address, false))
    return 0;

  do
  {
    if (!wait_for_irq(1000000))
    {
      printf("write (no irq): status = %x\n", (uint32_t) _regs[I2c_status]);
      stop();
      return nwritten;
    }
  }
  while (data_valid() && !nack());

  while ((nwritten < length) && !nack())
  {
    _regs[I2c_data] = buf[nwritten++];
    send_next();

    do
    {
      if (!wait_for_irq(1000000))
      {
        printf("write (no irq): status = %x\n", (uint32_t) _regs[I2c_status]);
        stop();
        return nwritten;
      }
    }
    while (data_valid() && !nack());
  }

  stop();

  do
  {
    if (!wait_for_irq(1000000))
      break;
  }
  while (!transferred());

  return nwritten;
}

// Test for data validity.

bool
I2c_jz4730_channel::data_valid()
{
  return (_regs[I2c_status] & I2c_status_data_valid) ? true : false;
}

// Request the next byte by clearing the data validity flag.

void
I2c_jz4730_channel::clear_next()
{
  _regs[I2c_status] = _regs[I2c_status] & ~I2c_status_data_valid;
}

// Indicate data ready for sending.

void
I2c_jz4730_channel::send_next()
{
  _regs[I2c_status] = _regs[I2c_status] | I2c_status_data_valid;
}

// Test for non-acknowledgement.

bool
I2c_jz4730_channel::nack()
{
  return (_regs[I2c_status] & I2c_status_nack) ? true : false;
}

// Set non-acknowledgement when receiving data.

void
I2c_jz4730_channel::signal_last()
{
  _regs[I2c_control] = _regs[I2c_control] | I2c_control_nack;
}

// Test for bus activity.

bool
I2c_jz4730_channel::busy()
{
  return (_regs[I2c_status] & I2c_status_busy) ? true : false;
}

// Test for transfer activity.

bool
I2c_jz4730_channel::transferring()
{
  return (_regs[I2c_status] & I2c_status_buffer_nempty) ? true : false;
}

// Test for write transfer completion.

bool
I2c_jz4730_channel::transferred()
{
  return (_regs[I2c_status] & I2c_status_transmit_end) ? true : false;
}

// Explicitly start communication.

void
I2c_jz4730_channel::start()
{
  _regs[I2c_control] = (_regs[I2c_control] & ~I2c_control_nack) | I2c_control_start;
}

// Explicitly stop communication.

void
I2c_jz4730_channel::stop()
{
  _regs[I2c_control] = _regs[I2c_control] | I2c_control_stop;
}



// Initialise the I2C controller.

I2c_jz4730_chip::I2c_jz4730_chip(l4_addr_t start, l4_addr_t end,
                                 Cpm_jz4730_chip *cpm,
                                 uint32_t frequency)
: _start(start), _end(end), _cpm(cpm), _frequency(frequency)
{
}

// Obtain a channel object. Only one channel is supported.

I2c_jz4730_channel *
I2c_jz4730_chip::get_channel(uint8_t channel, l4_cap_idx_t irq)
{
  if (channel == 0)
    return new I2c_jz4730_channel(_start, _cpm, _frequency, irq);
  else
    throw -L4_EINVAL;
}



// C language interface functions.

void *jz4730_i2c_init(l4_addr_t start, l4_addr_t end, void *cpm, uint32_t frequency)
{
  return (void *) new I2c_jz4730_chip(start, end, static_cast<Cpm_jz4730_chip *>(cpm), frequency);
}

void *jz4730_i2c_get_channel(void *i2c, uint8_t channel, l4_cap_idx_t irq)
{
  return static_cast<I2c_jz4730_chip *>(i2c)->get_channel(channel, irq);
}

void jz4730_i2c_disable(void *i2c_channel)
{
  static_cast<I2c_jz4730_channel *>(i2c_channel)->disable();
}

void jz4730_i2c_enable(void *i2c_channel)
{
  static_cast<I2c_jz4730_channel *>(i2c_channel)->enable();
}

unsigned int jz4730_i2c_read(void *i2c_channel, uint8_t address, uint8_t buf[], unsigned int length)
{
  return static_cast<I2c_jz4730_channel *>(i2c_channel)->read(address, buf, length);
}

unsigned int jz4730_i2c_write(void *i2c_channel, uint8_t address, uint8_t buf[], unsigned int length)
{
  return static_cast<I2c_jz4730_channel *>(i2c_channel)->write(address, buf, length);
}
